(********* Matematica - corso di LPL: Laboratorio 2017 *************)
(*******************************************************************)
(* Definizione del Linguaggio SmallC: Nucleo di C esteso           *)
(* Realizzazione di un Esecutore di SmallC utilizzando OCaml       *)
(*******************************************************************)

(* =============================================================== *)
(* The Token, i.e. Lexical Categories *)

type ide = string;;
type num = int;;


(* =============================================================== *)
(* The Syntactic Categories *)

type dcl = 
	Const of ide * num
	| Var of ide * num
	| VarN of ide
	| Array of ide * num
	| SeqDcl of dcl * dcl
	| EmptyDcl
	;;
	
type exp = 
	Val of ide
	| Num of num
	| GetArrow of ide * exp
	| Plus of exp * exp
	| Minus of exp * exp
	| Mul of exp * exp
	| Div of exp * exp
	| Eq of exp * exp
	| LT of exp * exp
	| GT of exp * exp
	| Not of exp
	| Or of exp * exp
	| And of exp * exp
	;;
	
type cmd =
	UpdVar of ide * exp
	| UpArray of ide * exp * exp
	| Ifte of exp * cmd * cmd
	| Ift of exp * cmd
	| While of exp * cmd
	| SeqCmd of cmd * cmd
	;;
	
type prog = Program of dcl * cmd | ProgN of cmd;;


(* =============================================================== *)
(* The Semantic Structure: Store e Environment *)	

(* Store: Structure *)
type loc = Loc of int;;
type mval = Mval of int
	| Undef;;

type store = Store of int * (loc -> mval);;

(* Store: Operations and Exceptions *)
let storeSize = 1000;;

exception IllegalAddress of string * loc;;
exception UndefinedLoc of string * loc;;
exception StoreSizeExceeded of string * int * int;;

let emptyStore () =  
		Store (0, fun (l:loc) -> Undef)
		;;
let allocate (Store(d,g)) k = 
		if (d+k<=storeSize)
		then ((Loc d),Store(d+k,g))
		else raise (StoreSizeExceeded ("allocate",d,k))
		;;
let upd (Store(d,g)) loc mv =
		let checkLoc = fun (Loc x) -> x >= 0 && x < d
		in	if checkLoc loc
			then Store(d, fun l -> if l = loc then mv else g l)
			else raise (IllegalAddress("upd",loc))
			;;
let checkLoc = fun (Loc x) d -> x >= 0 && x < d ;; 		
let getStore (Store(d,g)) loc =
		if not(checkLoc loc d) 
		then raise (IllegalAddress("getStore",loc))
		else match g loc with
				Undef -> raise (UndefinedLoc ("getStore",loc))
				| ret -> ret
				;;
						
(* Environment *)
type dval = 
	DConst of num
	| DVar of loc
	| DArray of loc
	| Unbound
	;;
	 
type env = Env of (ide -> dval);;

(* Environment: Operations and Exceptions *)

exception UnboundIde of string * ide;;

let emptyEnv () = 
		Env (fun ide -> Unbound)
		;;
let bind (Env r) ide dval =
		Env(fun i -> if i = ide then dval else r i)
		;;		
let getEnv (Env r) ide =
		match r ide with
			Unbound -> raise (UnboundIde ("getEnv",ide))
		    | ret -> ret
		   	;;

(* =============================================================== *)
(* The Semantic Function: dclSem, expSem, cmdSem, progSem *)

(* =============================================================== *)
(* The Semantic Function: dclSem *)

let rec dclSem dcl (rho,(Store(d,g)as mu)) = 
	match dcl with
		Const(ide,num) -> (bind rho ide (DConst num) , mu)
		| Var(ide,num) -> let (loc,mu') = allocate mu 1
						  in (bind rho ide (DVar loc) , upd mu' loc (Mval num))
		| VarN ide -> let (loc,mu') = allocate mu 1
						  in (bind rho ide (DVar loc) , mu')
	    | Array(ide,num) -> let ((Loc b)as loc,mu') = allocate mu num
	    				  	   in (bind rho ide (DArray loc) , mu')
	    | SeqDcl(dcl1,dcl2) -> let (rho',mu') = dclSem dcl1 (rho,mu)
	    					   in dclSem dcl2 (rho',mu')
		| _ -> (rho,mu)
		;;

(* =============================================================== *)
(* The Semantic Function: expSem *)
 		
exception UnInitialized of string * ide;;
exception TypeMismatch of string * ide * string;;

let rec expSem exp ((rho,mu)as sigma) =
	match exp with
		Val ide -> (match (getEnv rho ide) with
				   	  DConst num -> num
				   	  | DVar loc -> (match (getStore mu loc) with
				   					   Mval n -> n 
				   					   | _ -> raise (UnInitialized ("expSem",ide)))
				   	  | DArray loc -> raise (TypeMismatch ("expSem",ide,
				   	  	       "Wrong: "^ide^" is array of int but is used as scalar"))
				   	  | _ -> raise (UnboundIde ("expSem",ide)))
		| Num n -> n
		| GetArrow(ide,e) -> (match (getEnv rho ide) with
					DArray (Loc n) -> (match getStore mu (Loc(n+(expSem e sigma))) with
										Mval n -> n
										| _ -> raise (UnInitialized ("expSem",ide)))
					| DConst num -> raise (TypeMismatch("expSem",ide,
				   	  	"Error: "^ide^" is a constant but is used as an array of int"))
					| DVar loc -> raise (TypeMismatch("expSem",ide,
				   	  	"Error: "^ide^" is an int var but is used as an array of int"))
				   	| _ -> raise (UnboundIde ("expSem",ide)))
		| Plus(e1,e2) -> expSem e1 sigma + expSem e2 sigma
		| Minus(e1,e2) -> expSem e1 sigma - expSem e2 sigma
		| Mul(e1,e2) -> expSem e1 sigma * expSem e2 sigma
		| Div(e1,e2) -> expSem e1 sigma / expSem e2 sigma
		| Eq(e1,e2) -> if expSem e1 sigma = expSem e2 sigma then 1 else 0
		| LT(e1,e2) -> if expSem e1 sigma < expSem e2 sigma then 1 else 0
		| GT(e1,e2) -> if expSem e1 sigma > expSem e2 sigma then 1 else 0
		| Not e -> if expSem e sigma = 0 then 1 else 0
		| Or(e1,e2) -> if expSem e1 sigma = 0 || expSem e2 sigma = 0 then 0 else 1
		| And(e1,e2) -> if expSem e1 sigma != 0 && expSem e2 sigma != 0 then 1 else 0
		;;

(* ============================================================================ *)
(* The Semantic Function: cmdSem *)

let rec cmdSem cmd ((rho,mu)as sigma) =
	match cmd with
		UpdVar(ide,exp) -> (match (getEnv rho ide) with
				   	  DVar loc -> (rho,upd mu loc (Mval (expSem exp sigma)))
				   	  | _ -> raise (TypeMismatch ("cmdSem",ide,
				   	  	       "Wrong: "^ide^" is used as var but it isnt")))
		| UpArray(ide,exp1,exp2) -> (match (getEnv rho ide) with
					DArray (Loc n) -> let k = expSem exp1 sigma
										and v = expSem exp2 sigma
									  in (rho, upd mu (Loc (n+k)) (Mval v))
				   	  | _ -> raise (TypeMismatch ("cmdSem",ide,
				   	  	       "Wrong: "^ide^" is used as array but it isnt")))
		| Ifte(exp,c1,c2) -> (match (expSem exp sigma) with
					0 -> cmdSem c1 sigma
				   	  | _ -> cmdSem c2 sigma)
		| Ift(exp,c) -> (match (expSem exp sigma) with
					0 -> sigma
				   	  | _ -> cmdSem c sigma)
		| While(exp,c) -> (match (expSem exp sigma) with
					0 -> sigma
				   	  | _ -> cmdSem cmd (cmdSem c sigma))
		| SeqCmd(c1,c2) -> cmdSem c2 (cmdSem c1 sigma)
		;;

(* ============================================================================ *)
(* The Semantic Function: progSem *)

let rec progSem prog sigma =
	match prog with 
		Program(dcl,cmd) -> let sigma1 = dclSem dcl sigma
							in cmdSem cmd sigma1
		| ProgN cmd -> cmdSem cmd sigma
		;;





